/*
 * DeskLink for *nix (dl)
 * Copyright (C) 2004
 * Stephen Hurd
 *
 * Redistribution of modified and unmodified copies
 * is premitted provided the copyright remains intact
 */

/*
DeskLink+
Extensions and enhancements Copyright (C) 2005 John R. Hogerhuis

DeskLink+ is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License version 2 or any
later as version as published by the Free Software Foundation.  

DeskLink+ is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (in the file "COPYING"); if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111, USA.
*/


#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "dir_list.h"

#ifdef __darwin__
#include <util.h>
#endif

#if defined(__FreeBSD__)
#include <libutil.h>
#endif

#if defined(__NetBSD__) || defined(OpenBSD)
#include <util.h>
#endif

#ifdef __linux__
#include <utmp.h>
#include <netinet/in.h>
#endif


unsigned prev_dir_type = 0;
unsigned fname_ndx = 0;
unsigned file_len;
LocalID cur_id;

int file=-1;
int mode=0;	/* 0=unopened, 1=Write, 3=Read, 2=Append */
int term=-1;
unsigned char filename[25];
unsigned char dirp;
struct dirent *dire;
DIR *dir=NULL;
char *dirname;
char **args;
struct termios origt;
struct termios	ti;
int gettymode;
int guimode=1;
int debug=0;
int upcase=1;
unsigned dot_offset = 6; // standard for M100, NEC
unsigned char buf[131];

int bedisk(void);

void out_buf(unsigned char *bufp, unsigned len);

int my_write (int fh, void *srcp, size_t len)
{
   if (debug)
   {
      fprintf (stderr, "response: ");
      out_buf (srcp, len);
      fprintf (stderr, "\n");
   }
   return (write (fh, srcp, len));
}

unsigned char calc_sum(unsigned char type, unsigned char length, unsigned char *data)
{
	unsigned short sum=0;
	int i;

	sum+=type;
	sum+=length;
	for(i=0;i<length;i++)
		sum+=data[i];
	return((sum & 0xFF) ^ 255);
}

void normal_return(unsigned char type)
{
	buf[0]=0x12;
	buf[1]=0x01;
	buf[2]=type;
	buf[3]=calc_sum(0x12,0x01,buf+2);
	my_write(term,buf,4);
	if(debug)
		fprintf(stderr, "Response: %02X\n",type);
}

int main(int argc, char **argv)
{
    int off=0;
    unsigned char *p;
    unsigned char path[PATH_MAX];
    int arg;

    /** create the file list (for reverse order traversal) */
    file_list_init ();

    args = argv;
    dirname = (char *)strdup((char *)(argv[0]));
    p = (unsigned char *) strrchr ((char *)dirname, '/');
    if (p != NULL)
    {
        *p = 0;
        chdir ((char *)dirname);
    }

    strcpy ((char *)path, "/dev/ttyS0");
    for (arg = 1; arg < argc; arg++)
    {
        switch (argv[arg][0])
        {
            case '/':
                strcpy ((char *)path, (char *)(argv[arg]));
                break;

            case '-':
                switch (argv [arg][1]) {

                    case 0:
                        strcpy ((char *)path,"/dev/tty");
                        term = 1;
                        break;

                    case 'g':

                        gettymode=1;
                        guimode=0;
                        break;

                    case 'f':

                        upcase = 0;

                    case 'c':

                        guimode=0;
                        break;

                    case 'v':

                        debug=1;
                        break;


                    case 'p':

                        /** set the current directory */
                        if (argv[arg][2] != '=') break;
                        chdir (argv[arg] + 3);
                        system ("pwd");

                        break;

                    case 'w':

                        /** set the 'dot' offset to match wp-2 */
                        dot_offset = 8;

                        break;

                    case 'h':
print_usage:
                        fprintf (stderr, "dl [dev name] [options]\n");
                        fprintf (stderr, "   -v       Verbose (debug) mode\n");
                        fprintf (stderr, "   -p=dir   Path to files to be served\n");
                        fprintf (stderr, "   -w       WP-2 compatibility mode\n");
                        fprintf (stderr, "   -f       Don't upcase file names when enumerating\n");
                        fprintf (stderr, "Example:\n");
                        fprintf (stderr, "   dl /dev/ttyUSB0 -p=/home/john/wp2files -w -v\n");

                        fprintf (stderr, "\n");

                        break;

                    default:
                        fprintf(stderr, "Unknown option %s\n",argv[arg]);
                        goto print_usage;
                }
                break;

            default:
                strcpy((char *)path,"/dev/");
                strcat((char *)path,(char *)(argv[arg]));
        }
    }

    if(term<0)
        term=open((char *)path,O_RDWR|O_NONBLOCK);
    if(term<0)
        return(1);

    if(gettymode) {
        if(login_tty(term)==0)
            term=STDIN_FILENO;
        else
            daemon(1,1);
    }

    (void)tcflush(term, TCIOFLUSH);	/* clear out the crap */
    ioctl(term, FIONBIO, &off);	/* turn off non-blocking mode */
    ioctl(term, FIOASYNC, &off);	/* ditto for async mode */

    if(tcgetattr(term,&ti)==-1)
        return(1);
    cfmakeraw(&ti);
    ti.c_cflag |= CLOCAL|CRTSCTS|CS8;
    if(cfsetspeed(&ti,B19200)==-1)
        return(1);
    if(tcsetattr(term,TCSANOW,&ti)==-1)
        return(1);

    /* Set up terminal termios struct */
    origt.c_iflag=BRKINT | ICRNL | IMAXBEL | IXON | IXANY | IXOFF;
    origt.c_oflag=OPOST | ONLCR;
    origt.c_lflag=ECHO | ICANON | ISIG | IEXTEN | ECHOE | ECHOKE | ECHOCTL;
    origt.c_cflag=CREAD | CS8 | HUPCL
#ifdef CRTS_OFLOW
        | CRTS_OFLOW
#else
        | CRTSCTS
#endif
#ifdef CRTS_IFLOW
        | CRTS_IFLOW
#endif
        ;

    cfsetspeed(&origt,B1200);

    while(1)
        bedisk();

    file_list_cleanup ();

    return(0);
}

int out_dirent (unsigned char *fnamep, unsigned len)
{
   unsigned short size;
   unsigned char *dotp;
   unsigned char *p;
   unsigned i;


   if (debug) fprintf (stderr, "out_dirent: %s\n", fnamep);
   
   /** format the filename */
   if (fnamep)
   {

       if (debug)
           fprintf (stderr, "no fmt: %s\n", fnamep);
       buf[26] = 'F';
       size = htons (len);
       memcpy (buf + 27, &size, 2);
       memset (buf + 2, ' ', 24);
       if (upcase)
       {
           for (i = 0; i < MIN (strlen ((char *)fnamep), 24); i++)
           {
               buf[2+i] = toupper (fnamep[i]);
           }
       }
       else
       {
           memcpy (buf + 2, fnamep, MIN (strlen ((char *)fnamep), 24));
       }
       dotp = memchr (buf + 2, '.', 24);
       if (dotp != NULL)
       {
           memmove (buf + dot_offset + 2, dotp, 3);
           for (p = dotp; p < buf + dot_offset + 2; p++)
               *p = ' ';
       }
       if (debug)
           fprintf (stderr, "str: %24.24s\n", (char *)buf + 2);
   }
   
   /** add checksum */
   buf[30] = calc_sum (0x11, 0x1C, buf + 2);

   /** write packet */
   return (my_write (term,buf,31) == 31);
}

int send_dirent (unsigned char *buf, struct stat *st)
{
	unsigned short size;
	unsigned char *dot;
	unsigned char *p;
    unsigned i;

	if(dire!=NULL) {
		buf[26]='F';
		size=htons(st->st_size);
		memcpy(buf+27,&size,2);
		memset(buf+2,' ',24);
        if (upcase)
        {
            for (i = 0; i < strlen (dire->d_name); i++)
            {
                buf[2+i] = toupper (dire->d_name[i]);
            }
        }
        else
        {
            memcpy(buf+2,dire->d_name,strlen((char *)dire->d_name));
        }
		dot=memchr(buf+2,'.',24);
		if(dot!=NULL) {
			memcpy(buf + dot_offset + 2, dot, 3);
			for( p = dot; p < buf + dot_offset + 2; p++)
				*p=' ';
		}
	}
	buf[30] = calc_sum (0x11, 0x1C, buf+2);
	return (my_write (term, buf, 31) == 31);
}

int read_next_dirent(struct stat *st)
{

    if (dir == NULL)
    {
    printf ("%s:%u\n", __FUNCTION__, __LINE__);
        dire=NULL;
        normal_return (0x70);
        return(0);
    }
    while((dire=readdir(dir)) != NULL)
    {
        if (debug)
        {
            fprintf (stderr, "%s (%u)", dire->d_name, (unsigned) st->st_size);
        }

        if(stat(dire->d_name,st))
        {
            normal_return(0x31);
            return 0;
        }

        if (!S_ISREG (st->st_mode))
        {
            continue;
        }

        if(st->st_size > 65535)
        {
            continue;
        }

        /** add file to list so we can traverse any order */
        add_file ((unsigned char *) dire->d_name, st->st_size, ++fname_ndx);
        if (debug)
            fprintf (stderr, "added file %s len %ld\n", dire->d_name, st->st_size);

        break;
    }
    if (dire == NULL) return 0;
    return 1;
}


int directory(unsigned char length, unsigned char *data)
{
	unsigned char search_form;
	unsigned char *p,*dot;
	struct stat st;

	bzero(buf,31);
	buf[29]=40;
	buf[0]=0x11;
	buf[1]=0x1C;
	search_form=data[length-1];
	switch (search_form)
    {

        case 0x00:	/* Pick file for open/delete */
            strncpy((char *)filename,(char *)data,24);
            filename[25]=0;
            /* Remove trailing spaces */
            for(p = (unsigned char *) strrchr((char *)filename,' '); p >= filename && *p == ' '; p--)
            {
                *p = 0;
            }
            /* Remove spaces between base and dot */
            dot = (unsigned char *) strchr((char *)filename,'.');
            if(dot!=NULL) {
                for(p=dot-1;*p==' ';p--);
                memmove(p+1,dot,strlen((char *)dot)+1);
            }
            if(dir!=NULL)
                closedir(dir);

            dir=opendir(".");
            if (upcase)
            {
                for(;read_next_dirent(&st) && dire!=NULL && strcasecmp((char *)dire->d_name, (char *)filename););
            }
            else
            {
                for(;read_next_dirent(&st) && dire!=NULL && strcmp((char *)dire->d_name, (char *)filename););
            }
            send_dirent(buf,&st);
            break;

        case 0x01:	/* "first" directory block */
            if (debug)
            {
                fprintf (stderr, "get first directory entry command\n");
            }
            
            if(dir!=NULL)
                closedir(dir);
            dir=opendir(".");

            /** rebuild the file list */
            file_list_clear_all();
            while (read_next_dirent (&st)) ;

            /** send the file name */
            if (get_first_file (filename, &file_len, &cur_id) == 0)
            {
                if (debug)
                    fprintf (stderr, "get_first_file -> %s len = %d\n", filename, file_len);
                out_dirent (filename, file_len);
            }
            else
                out_dirent (NULL, 0);

            break;

        case 0x02:	/* "next" directory block */

            if (get_next_file (filename, &file_len, &cur_id) == 0)
                out_dirent (filename, file_len);
            else
                out_dirent (NULL, 0);

            break;

        case 0x03:	/* "previous" directory block */

            if (get_prev_file (filename, &file_len, &cur_id) == 0)
                out_dirent (filename, file_len);
            else
                out_dirent (NULL, 0);
            break;

        case 0x04:	/* end directory reference */
            closedir(dir);
            file_list_clear_all ();
            fname_ndx = 0;
            break;
    }
	return 0;
}

int open_file(unsigned char omode)
{
	struct stat st;

	if(debug)
		fprintf (stderr, "Open Mode: %d\n", omode);
	switch(omode) {
		case 0x01:	/* New file for my_write */
			if (file >= 0) 
            {
				close(file);
				file=-1;
			}
			file = open ((char *) filename,O_CREAT|O_TRUNC|O_WRONLY|O_EXCL,0666);
			if(file<0)
				normal_return(0x37);
			else
            {
				mode=omode;
				normal_return(0x00);
			}
			break;
		case 0x02:	/* existing file for append */
			if (file >= 0)
            {
				close(file);
				file=-1;
			}
			stat ((char *) filename, &st);
			if(st.st_size>65535)
            {
				normal_return (0x6E);
				break;
			}
			file = open ((char *) filename, O_WRONLY | O_APPEND);
			if (file < 0)
				normal_return (0x37);
			else
            {
				mode=omode;
				normal_return (0x00);
			}
			break;
		case 0x03:	/* Existing file for read */
			if (file >= 0)
            {
				close (file);
				file=-1;
			}
			stat ((char *)filename, &st);
			if (st.st_size > 65535)
            {
				normal_return (0x6E);
				break;
			}
			file = open ((char *)filename, O_RDONLY);
			if(file<0)
				normal_return(0x37);
			else
            {
                mode = omode;
                normal_return (0x00);
            }
			break;
	}
	return (file);
}

void read_file(void)
{
	int in;

	buf[0]=0x10;
	if(file<0)
    {
		normal_return(0x30);
		return;
	}
	if(mode!=3)
    {
		normal_return(0x37);
		return;
	}
	in = read (file, buf+2, 128);
	buf[1] = (unsigned char) in;
	buf[2+in] = calc_sum(0x10, (unsigned char)in, buf+2);
	my_write (term, buf, 3+in);
}

void respond_mystery()
{
   static unsigned char canned[] =
      {0x38, 0x01, 0x00};

   memcpy (buf, canned, sizeof (canned));
   buf[sizeof(canned)] = calc_sum (canned[0], canned[1], canned + 2);
   my_write (term, buf, sizeof (canned) + 1);
}

void respond_mystery2()
{
   static unsigned char canned[] =
      {0x14, 0x0F, 0x41, 0x10, 0x01,
       0x00, 0x50, 0x05, 0x00, 0x02,
       0x00, 0x28, 0x00, 0xE1, 0x00,
       0x00, 0x00};

   memcpy (buf, canned, sizeof (canned));
   buf[sizeof(canned)] = calc_sum (canned[0], canned[1], canned + 2);
   my_write (term, buf, sizeof (canned) + 1);
}

void respond_place_path()
{
   static unsigned char canned[] =
    {0x12, 0x0b, 0x00, 0x52, 0x4f, 0x4f, 0x54, 0x20, 0x20, 0x2e, 0x3c, 0x3e, 0x20, 0x96};

   my_write (term, canned, sizeof (canned));
}

void renamefile(unsigned char *name)
{
	unsigned char *p;

	name[24]=0;
	for(p = (unsigned char *) strrchr((char *)name,' ');*p==' ';p--)
		*p=0;
	if(rename ((char *) filename, (char *)name))
		normal_return(0x4A);
	else
		normal_return(0x00);
}

int readbytes(int handle, void *buf, int max)
{
	int r = 0;
   int rval;
   unsigned i;

	while (r < max)
   {
      rval = read (term, buf + r, 1);
      if (rval < 0) continue;
		r += rval;
   }

   if (debug)
      for (i = 0; r >=0 && i < r; i++)
         fprintf (stderr, "%02X ", ((unsigned char *)buf)[i]);
   
	return (r);
}

int sendloader(void)
{
	int w=0;
	int i=0;
	int fd;
	unsigned char b;

	if((fd=open("loader.ba",O_RDONLY))<0) {
		if(debug)
			fprintf(stderr, "Failed to open loader.ba for read.\n");
		return(9);
	}

	sleep(1);

	while(read(fd,&b,1)==1) {
		while((i=my_write(term,&b,1))!=1);
		w+=i;
		usleep(5000);
		if(debug) {
			fprintf(stderr, "\rSent: %d bytes",w);
			fflush(stdout);
		}
	}
	b=0x1A;
	my_write(term,&b,1);
	close(fd);
	close(term);
	if(debug) {
		puts("");
		if(strlen((char *)(args[0]))>4)
			strcpy((char*) (args[0]),"./dl");
		execvp("./dl",args);
	}
	else {
		exit(0);
	}

	return(0);
}

void out_buf(unsigned char *bufp, unsigned len)
{
   unsigned i,j,k;

   if (!debug) return;
   
   for (i = 0; i < len;)
   {
      for (j = 0; j < 2; j++)
      {
         for (k = 0; i < len && k < 8; k++, i++)
            fprintf (stderr, "%02X ", bufp[i]);
         fprintf (stderr, "   ");
      }
      fprintf (stderr, "\n");
   }
   
   fprintf (stderr, "\n");

}


int bedisk(void)
{
    unsigned char preamble[3];
    unsigned char type;
    unsigned char length;
    unsigned char data[0x81];
    unsigned char checksum;
    unsigned precnt = 0;
    unsigned len;

    preamble[2]=0;

    for (precnt = 0; precnt < 2; precnt++)
    {
        len = readbytes(term, preamble + precnt, 1);
        if (len == 0)
        {
            fprintf (stderr, "zero length - 1\n");
            continue;
        }
        if (preamble[precnt] != 'Z' && preamble[precnt] != 'Y' && preamble[precnt] != 'X')
            return (-1); // no error message... TS-DOS doesn't like that...
    }

    if(preamble[0]==preamble[1]) {
        switch(preamble[0]) {
            case 'Z':
                break;
            case 'Y':
                tcsetattr(term,TCSANOW,&origt);
                system("login");
                tcsetattr(term,TCSANOW,&ti);
                return(10);
            case 'X':
                sendloader();
                return(sendloader());
            default:
                if(debug)
                    fprintf(stderr, "Bad preamble: '%s' (%02X %02X)\n",preamble,preamble[0],preamble[1]);
                normal_return(0x40);
                return(2);
        }
    }
    else {
        if(debug)
            fprintf(stderr, "Bad preamble: '%s' (%02X %02X)\n",preamble,preamble[0],preamble[1]);
    }


    len = readbytes(term,&type,1);
    if (!len)
    {
        fprintf (stderr, "zero len - 2\n");
        return (0);
    }

    len = readbytes(term,&length,1);
    if (!len)
    {
        fprintf (stderr, "zero len - 3\n");
        return (0);
    }

    len = readbytes(term,data,length);
    if (length && !len)
    {
        fprintf (stderr, "zero len - 4\n");
        return (0);
    }
    data[length]=0;

    len = readbytes(term,&checksum,1);
    if (!len)
    {
        fprintf (stderr, "zero len - 5\n");
        return (0);
    }

    if (debug)
        fprintf (stderr, "\nProcessing command type = %02X length = %02X csum = %02X\n", type, length, checksum);

    if(calc_sum(type,length,data)!=checksum) {
        if(debug) {
            fprintf(stderr, "BAD CHECKSUM! Preamble: %s Type: %02X Length: %02X Data: %s\n",preamble,type,length,data);
            fprintf(stderr, "Packet checksum: %02X  My checksum: %02X\n",checksum,calc_sum(type,length,data));
        }
        normal_return(0x36);
        return(7);
    }

    switch(type) {
        case 0x00:	/* Directory ref */
            directory(length,data);
            break;
        case 0x01:	/* Open file */
            open_file(data[0]);
            break;
        case 0x02:	/* Close file */
            if(file>=0)
                close(file);
            normal_return(0x00);
            break;
        case 0x03:	/* Read */
            read_file();
            break;
        case 0x04:	/* Write */
            if(file<0) {
                normal_return(0x30);
                break;
            }
            if(mode!=1 && mode !=2) {
                normal_return(0x37);
                break;
            }
            if(my_write(file,data,length)!=length)
                normal_return(0x4a);
            else
                normal_return(0x00);
            break;
        case 0x05:	/* Delete */
            unlink ((char *) filename);
            normal_return(0x00);
            break;
        case 0x06:	/* Format disk */
            normal_return(0x00);
            break;
        case 0x07:	/* Drive Status */
            normal_return(0x00);
            break;
        case 0x08:	/* Drive condition */
            respond_place_path ();
            break;
        case 0x0D:	/* Rename File */
            renamefile(data);
            break;
        case 0x23:  /* TS-DOS mystery command 2 */
            respond_mystery2();
            break;
        case 0x31:  /* TS-DOS mystery command 1 */
            respond_mystery();
            break;
        default:
            return(8);
            break;
    }
    return(0);
}
